﻿EnableExplicit

#GENERATE_STACKMACHINE_ASM = 1

XIncludeFile "Evaluate.pbi"

Define mainWindow
Define inputGadget, resultGadget, newVar, addVar
Define gadgetFont, eventType
Define t1,t2,t3,t4,t5

Global errorOut, asmOut, variablesView

Procedure myErrorHandler(error.l,arg1.s)
    Define msg.s
    
    Select error
        Case #evalError_HexNumExpected
            msg="ERROR: expected a HEX number after $"
        Case #evalError_BinNumExpected
            msg="ERROR: expected a binary number after %"
        Case #evalError_IllegalInput
            msg="ERROR: illegal input detected: '" + arg1 +"' - (Chr: "+Str(Asc(arg1))+")"
        Case #evalError_MissingOperator
            msg="ERROR: operator expected"
        Case #evalError_MissingOperand
            msg="ERROR: missing operand after: " +arg1
        Case #evalError_MissingExpression
            msg="ERROR: expression expected after " + arg1
        Case #evalError_MissingRParenthesis
            msg="ERROR: missing ')'"
        Case #evalError_MissingDecimalPlace
            msg="ERROR: missing decimal place after ."
        Case #evalError_BinaryNotWithFloats
            msg="ERROR: binary "+arg1+" with floating point values not supported, use integer"
        Case #evalError_DivisionWithNull
            msg="ERROR: division with 0"
        Case #evalError_ModuloWithNull
            msg="ERROR: modulo (%) with 0"
        Case #evalError_WrongFloatNumberFormat
            msg="ERROR: wrong floating point number format"
        Case #evalError_TokenCantStartExpression
            msg="ERROR: "+arg1+" can't start an expression"
        Default
            msg="ERROR: unknown error: "+arg1
    EndSelect
    AddGadgetItem(errorOut,-1,msg)
EndProcedure

Procedure myAsmHandler(txt.s)
    AddGadgetItem(asmOut,-1,txt)
EndProcedure

Procedure updateVariablesView()
    ClearGadgetItems(variablesView)
    ForEach EvaluateVariables()
        AddGadgetItem(variablesView,-1,MapKey(EvaluateVariables())+Chr(10)+EvaluateVariables() )
    Next
EndProcedure


SetEvaluateErrorHandler(@myErrorHandler())

SetEvaluateAsmOutputHandler(@myAsmHandler())

gadgetFont = LoadFont(#PB_Any,"Lucida Console",12)

mainWindow = OpenWindow(#PB_Any,0,0,1024,600,"Evaluate()",#PB_Window_SystemMenu|#PB_Window_ScreenCentered)
If mainWindow
    SetWindowColor(mainWindow,RGB($50,$50,$50))

    inputGadget = EditorGadget(#PB_Any,300,20,724,100)
        SetGadgetColor(inputGadget,#PB_Gadget_BackColor ,RGB($30,$30,$30))
        SetGadgetColor(inputGadget,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))
        SetGadgetFont(inputGadget,FontID(gadgetFont))

    resultGadget = StringGadget(#PB_Any,300,140,724,30,"",#PB_String_ReadOnly)
        SetGadgetColor(resultGadget,#PB_Gadget_BackColor ,RGB($30,$30,$30))
        SetGadgetColor(resultGadget,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))
        SetGadgetFont(resultGadget,FontID(gadgetFont))

    errorOut = EditorGadget(#PB_Any,300,190,724,100,#PB_Editor_ReadOnly)
        SetGadgetColor(errorOut,#PB_Gadget_BackColor ,RGB($00,$00,$00))
        SetGadgetColor(errorOut,#PB_Gadget_FrontColor,RGB($FF,$00,$00))
        SetGadgetFont(errorOut,FontID(gadgetFont))
    
    asmOut   = ListViewGadget(#PB_Any,300,310,724,290)
        SetGadgetColor(asmOut,#PB_Gadget_BackColor ,RGB($00,$00,$00))
        SetGadgetColor(asmOut,#PB_Gadget_FrontColor,RGB($00,$FF,$80))
        SetGadgetFont(asmOut,FontID(gadgetFont))
        
    variablesView = ListIconGadget(#PB_Any,0,20,280,500,"Name",100,#PB_ListIcon_FullRowSelect|#PB_ListIcon_GridLines)
        AddGadgetColumn(variablesView,1,"Value",170)
        SetGadgetColor(variablesView,#PB_Gadget_BackColor ,RGB($00,$00,$00))
        SetGadgetColor(variablesView,#PB_Gadget_FrontColor,RGB($00,$FF,$80))
        SetGadgetColor(variablesView,#PB_Gadget_LineColor ,RGB($30,$30,$30))
        SetGadgetFont(variablesView,FontID(gadgetFont))

    newVar = StringGadget(#PB_Any,0,520,280,35,"n = 1.2345")
        SetGadgetColor(newVar,#PB_Gadget_BackColor ,RGB($30,$30,$30))
        SetGadgetColor(newVar,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))
        SetGadgetFont(newVar,FontID(gadgetFont))
    addVar = ButtonGadget(#PB_Any,0,555,280,35,"Add Variable")

    t1=TextGadget(#PB_Any,  0,  0,280,20,"Variables" ,#PB_Text_Center)
        SetGadgetColor(t1,#PB_Gadget_BackColor ,RGB($50,$50,$50))
        SetGadgetColor(t1,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))
    t2=TextGadget(#PB_Any,300,  0,724,20,"Input"     ,#PB_Text_Center)
        SetGadgetColor(t2,#PB_Gadget_BackColor ,RGB($50,$50,$50))
        SetGadgetColor(t2,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))
    t3=TextGadget(#PB_Any,300,120,724,20,"Result"    ,#PB_Text_Center)
        SetGadgetColor(t3,#PB_Gadget_BackColor ,RGB($50,$50,$50))
        SetGadgetColor(t3,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))
    t4=TextGadget(#PB_Any,300,170,724,20,"Errors"    ,#PB_Text_Center)
        SetGadgetColor(t4,#PB_Gadget_BackColor ,RGB($50,$50,$50))
        SetGadgetColor(t4,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))
    t5=TextGadget(#PB_Any,300,290,724,20,"ASM Output",#PB_Text_Center)
        SetGadgetColor(t5,#PB_Gadget_BackColor ,RGB($50,$50,$50))
        SetGadgetColor(t5,#PB_Gadget_FrontColor,RGB($AA,$AA,$AA))

    EvaluateVariables("pi") = StrD(#PI,1000)
    updateVariablesView()

    SetGadgetText(inputGadget,"12 + 54 * 7 /"+#CRLF$+"3 * 0.5 - 0.5 +"+#CRLF$+"12*(1+3*(1+1)) - 123"+#CRLF$)

    SetActiveGadget(inputGadget)

    Repeat
        Select WaitWindowEvent()
            Case #PB_Event_Gadget
                eventType = EventType()
                Select EventGadget()
                    Case inputGadget
                        Define text.s, i
                        text=""
                        ClearGadgetItems(asmOut)
                        ClearGadgetItems(errorOut)
                        For i = 0 To CountGadgetItems(inputGadget)
                           text + GetGadgetItemText(inputGadget,i)
                        Next
                        text = Evaluate( text )
                        SetGadgetText(resultGadget,text)
                        If text="ERROR"
                            ClearGadgetItems(asmOut)
                        EndIf
                    Case addVar
                        Define varName.s, varValue.s
                        varName  = Trim(StringField( GetGadgetText(newVar),1,"="))
                        varValue = Trim(StringField( GetGadgetText(newVar),2,"="))
                        If FindString(varValue,".",1) Or FindString(UCase(varValue),"E",1)
                            ;varValue = StrD(ValD(varValue),1000)
                        Else
                            varValue = Str(Val(varValue))
                        EndIf
                        EvaluateVariables(varName)=varValue
                        updateVariablesView()
                EndSelect
            Case #PB_Event_CloseWindow
                Break
        EndSelect
    ForEver
EndIf

; IDE Options = PureBasic 4.51 (Windows - x86)
; CursorPosition = 151
; FirstLine = 102
; Folding = -
; EnableUnicode
; EnableXP
; Executable = EvaluateExpression9.exe
; CPU = 5